@@ -7,10 +7,7 @@
 # Documentation
 ##############################################################################
 
-"""
-A library of subtree creators that build complex patterns of behaviours
-representing common behaviour tree idioms.
-"""
+"""Creators of common subtree patterns."""
 
 ##############################################################################
 # Imports
@@ -18,14 +15,8 @@ representing common behaviour tree idioms.
 
 import operator
 import typing
-import uuid
 
-from . import behaviour
-from . import behaviours
-from . import blackboard
-from . import common
-from . import composites
-from . import decorators
+from . import behaviour, behaviours, blackboard, common, composites, decorators
 
 ##############################################################################
 # Creational Methods
@@ -33,9 +24,11 @@ from . import decorators
 
 
 def pick_up_where_you_left_off(
-        name="Pickup Where You Left Off Idiom",
-        tasks=[]):
+    name: str, tasks: typing.List[behaviour.BehaviourSubClass]
+) -> behaviour.Behaviour:
     """
+    Create an idiom that enables a sequence of tasks to pick up where it left off.
+
     Rudely interrupted while enjoying a sandwich, a caveman (just because
     they wore loincloths does not mean they were not civilised), picks
     up his club and fends off the sabre-tooth tiger invading his sanctum
@@ -71,22 +64,23 @@ def pick_up_where_you_left_off(
     Returns:
         :class:`~py_trees.behaviour.Behaviour`: root of the generated subtree
     """
-    root = composites.Sequence(name=name)
+    root = composites.Sequence(name=name, memory=True)
     for task in tasks:
-        task_selector = composites.Selector(name="Do or Don't")
+        task_selector = composites.Selector(name="Do or Don't", memory=False)
         task_guard = behaviours.CheckBlackboardVariableValue(
             name="Done?",
             check=common.ComparisonExpression(
                 variable=task.name.lower().replace(" ", "_") + "_done",
                 value=True,
-                operator=operator.eq
-            )
+                operator=operator.eq,
+            ),
         )
-        sequence = composites.Sequence(name="Worker")
+        sequence = composites.Sequence(name="Worker", memory=True)
         mark_task_done = behaviours.SetBlackboardVariable(
             name="Mark\n" + task.name.lower().replace(" ", "_") + "_done",
             variable_name=task.name.lower().replace(" ", "_") + "_done",
-            variable_value=True
+            variable_value=True,
+            overwrite=True,
         )
         sequence.add_children([task, mark_task_done])
         task_selector.add_children([task_guard, sequence])
@@ -94,102 +88,21 @@ def pick_up_where_you_left_off(
     for task in tasks:
         clear_mark_done = behaviours.UnsetBlackboardVariable(
             name="Clear\n" + task.name.lower().replace(" ", "_") + "_done",
-            key=task.name.lower().replace(" ", "_") + "_done"
+            key=task.name.lower().replace(" ", "_") + "_done",
         )
         root.add_child(clear_mark_done)
     return root
 
 
-def eternal_guard(
-        subtree: behaviour.Behaviour,
-        name: str="Eternal Guard",
-        conditions: typing.List[behaviour.Behaviour]=[],
-        blackboard_namespace: str=None) -> behaviour.Behaviour:
-    """
-    The eternal guard idiom implements a stronger :term:`guard` than the typical check at the
-    beginning of a sequence of tasks. Here they guard continuously while the task sequence
-    is being executed. While executing, if any of the guards should update with
-    status :data:`~common.Status.FAILURE`, then the task sequence is terminated.
-
-    .. graphviz:: dot/idiom-eternal-guard.dot
-        :align: center
-
-    Args:
-        subtree: behaviour(s) that actually do the work
-        name: the name to use on the root behaviour of the idiom subtree
-        conditions: behaviours on which tasks are conditional
-        blackboard_namespace: applied to condition variable results stored on the blackboard (default: derived from the idiom name)
-
-    Returns:
-        the root of the idiom subtree
-
-    .. seealso:: :class:`py_trees.decorators.EternalGuard`
-    """
-    if blackboard_namespace is None:
-        blackboard_namespace = name.lower().replace(" ", "_")
-    blackboard_variable_names = []
-    # construct simple, easy to read, variable names (risk of conflict)
-    counter = 1
-    for condition in conditions:
-        suffix = "" if len(conditions) == 1 else "_{}".format(counter)
-        blackboard_variable_names.append(
-            blackboard.Blackboard.separator +
-            blackboard_namespace +
-            "_condition" +
-            suffix
-        )
-        counter += 1
-    # if there is just one blackboard name already on the blackboard, switch to unique names
-    conflict = False
-    for name in blackboard_variable_names:
-        try:
-            unused_name = blackboard.Blackboard.get(name)
-            conflict = True
-        except KeyError:
-            pass
-    if conflict:
-        blackboard_variable_names = []
-        counter = 1
-        unique_id = uuid.uuid4()
-        for condition in conditions:
-            suffix = "" if len(conditions) == 1 else "_{}".format(counter)
-            blackboard_variable_names.append(blackboard_namespace + "_" + str(unique_id) + "_condition" + suffix)
-            counter += 1
-    # build the tree
-    root = composites.Parallel(
-        name=name,
-        policy=common.ParallelPolicy.SuccessOnAll(synchronise=False)
-    )
-    guarded_tasks = composites.Selector(name="Guarded Tasks")
-    for condition, blackboard_variable_name in zip(conditions, blackboard_variable_names):
-        decorated_condition = decorators.StatusToBlackboard(
-            name="StatusToBB",
-            child=condition,
-            variable_name=blackboard_variable_name
-        )
-        root.add_child(decorated_condition)
-        guarded_tasks.add_child(
-            behaviours.CheckBlackboardVariableValue(
-                name="Abort on\n{}".format(condition.name),
-                check=common.ComparisonExpression(
-                    variable=blackboard_variable_name,
-                    value=common.Status.FAILURE,
-                    operator=operator.eq
-                )
-            )
-        )
-    guarded_tasks.add_child(subtree)
-    root.add_child(guarded_tasks)
-    return root
-
-
 def either_or(
     conditions: typing.List[common.ComparisonExpression],
     subtrees: typing.List[behaviour.Behaviour],
-    name="Either Or",
-    namespace: typing.Optional[str]=None
+    name: str = "Either Or",
+    namespace: typing.Optional[str] = None,
 ) -> behaviour.Behaviour:
     """
+    Create an idiom with selector-like qualities, but no priority concerns.
+
     Often you need a kind of selector that doesn't implement prioritisations, i.e.
     you would like different paths to be selected on a first-come, first-served basis.
 
@@ -243,31 +156,41 @@ def either_or(
     .. todo:: a version for which other subtrees can preempt (in an unprioritised manner) the active branch
     """
     if len(conditions) != len(subtrees):
-        raise ValueError("Must be the same number of conditions as subtrees [{} != {}]".format(
-            len(conditions), len(subtrees))
+        raise ValueError(
+            "Must be the same number of conditions as subtrees [{} != {}]".format(
+                len(conditions), len(subtrees)
+            )
         )
-    root = composites.Sequence(name=name)
-    configured_namespace: str = namespace if namespace is not None else \
-        blackboard.Blackboard.separator + name.lower().replace("-", "_").replace(" ", "_") + \
-        blackboard.Blackboard.separator + str(root.id).replace("-", "_").replace(" ", "_") + \
-        blackboard.Blackboard.separator + "conditions"
+    root = composites.Sequence(name=name, memory=True)
+    configured_namespace: str = (
+        namespace
+        if namespace is not None
+        else blackboard.Blackboard.separator
+        + name.lower().replace("-", "_").replace(" ", "_")
+        + blackboard.Blackboard.separator
+        + str(root.id).replace("-", "_").replace(" ", "_")
+        + blackboard.Blackboard.separator
+        + "conditions"
+    )
     xor = behaviours.CheckBlackboardVariableValues(
         name="XOR",
         checks=conditions,
         operator=operator.xor,
-        namespace=configured_namespace
+        namespace=configured_namespace,
     )
-    chooser = composites.Selector(name="Chooser")
+    chooser = composites.Selector(name="Chooser", memory=False)
     for counter in range(1, len(conditions) + 1):
-        sequence = composites.Sequence(name="Option {}".format(str(counter)))
-        variable_name = configured_namespace + blackboard.Blackboard.separator + str(counter)
+        sequence = composites.Sequence(
+            name="Option {}".format(str(counter)), memory=True
+        )
+        variable_name = (
+            configured_namespace + blackboard.Blackboard.separator + str(counter)
+        )
         disabled = behaviours.CheckBlackboardVariableValue(
             name="Enabled?",
             check=common.ComparisonExpression(
-                variable=variable_name,
-                value=True,
-                operator=operator.eq
-            )
+                variable=variable_name, value=True, operator=operator.eq
+            ),
         )
         sequence.add_children([disabled, subtrees[counter - 1]])
         chooser.add_child(sequence)
@@ -277,13 +200,14 @@ def either_or(
 
 def oneshot(
     behaviour: behaviour.Behaviour,
-    name: str="Oneshot",
-    variable_name: str="oneshot",
-    policy: common.OneShotPolicy=common.OneShotPolicy.ON_SUCCESSFUL_COMPLETION
+    name: str = "Oneshot",
+    variable_name: str = "oneshot",
+    policy: common.OneShotPolicy = common.OneShotPolicy.ON_SUCCESSFUL_COMPLETION,
 ) -> behaviour.Behaviour:
     """
-    Ensure that a particular pattern is executed through to
-    completion just once. Thereafter it will just rebound with the completion status.
+    Ensure that a particular pattern is executed through to completion just once.
+
+    Thereafter it will just rebound with the completion status.
 
     .. graphviz:: dot/oneshot.dot
 
@@ -304,54 +228,51 @@ def oneshot(
 
     .. seealso:: :class:`py_trees.decorators.OneShot`
     """
-    subtree_root = composites.Selector(name=name)
-    oneshot_with_guard = composites.Sequence(
-        name="Oneshot w/ Guard")
+    subtree_root = composites.Selector(name=name, memory=False)
+    oneshot_with_guard = composites.Sequence(name="Oneshot w/ Guard", memory=True)
     check_not_done = decorators.Inverter(
         name="Not Completed?",
         child=behaviours.CheckBlackboardVariableExists(
-            name="Completed?",
-            variable_name=variable_name
-        )
+            name="Completed?", variable_name=variable_name
+        ),
     )
     set_flag_on_success = behaviours.SetBlackboardVariable(
         name="Mark Done\n[SUCCESS]",
         variable_name=variable_name,
-        variable_value=common.Status.SUCCESS
+        variable_value=common.Status.SUCCESS,
+        overwrite=True,
     )
     # If it's a sequence, don't double-nest it in a redundant manner
     if isinstance(behaviour, composites.Sequence):
         behaviour.add_child(set_flag_on_success)
         sequence = behaviour
     else:
-        sequence = composites.Sequence(name="OneShot")
+        sequence = composites.Sequence(name="OneShot", memory=True)
         sequence.add_children([behaviour, set_flag_on_success])
 
     oneshot_with_guard.add_child(check_not_done)
     if policy == common.OneShotPolicy.ON_SUCCESSFUL_COMPLETION:
         oneshot_with_guard.add_child(sequence)
     else:  # ON_COMPLETION (SUCCESS || FAILURE)
-        oneshot_handler = composites.Selector(name="Oneshot Handler")
-        bookkeeping = composites.Sequence(name="Bookkeeping")
+        oneshot_handler = composites.Selector(name="Oneshot Handler", memory=False)
+        bookkeeping = composites.Sequence(name="Bookkeeping", memory=True)
         set_flag_on_failure = behaviours.SetBlackboardVariable(
             name="Mark Done\n[FAILURE]",
             variable_name=variable_name,
-            variable_value=common.Status.FAILURE
+            variable_value=common.Status.FAILURE,
+            overwrite=True,
         )
         bookkeeping.add_children(
-            [set_flag_on_failure,
-             behaviours.Failure(name="Failure")
-             ])
+            [set_flag_on_failure, behaviours.Failure(name="Failure")]
+        )
         oneshot_handler.add_children([sequence, bookkeeping])
         oneshot_with_guard.add_child(oneshot_handler)
 
     oneshot_result = behaviours.CheckBlackboardVariableValue(
         name="Oneshot Result",
         check=common.ComparisonExpression(
-            variable=variable_name,
-            value=common.Status.SUCCESS,
-            operator=operator.eq
-        )
+            variable=variable_name, value=common.Status.SUCCESS, operator=operator.eq
+        ),
     )
     subtree_root.add_children([oneshot_with_guard, oneshot_result])
     return subtree_root
