@@ -6,14 +6,12 @@
 # Documentation
 ##############################################################################
 
-"""
-Version number provided separately here so there is easy access for the module,
-setup.py and sphinx.
-"""
+"""Package version number."""
 
 ##############################################################################
 # Version
 ##############################################################################
 
 # When changing, Also update setup.py and package.xml
-__version__ = '2.1.5'
+# TODO: use pkg_resources to fetch the version from setup.py
+__version__ = "2.2.1"
